import os

MODELS_DIR = os.path.expanduser("~/.cache/wellcomeml/models")

MODEL_DISPATCH = {
    'scibert_scivocab_uncased': {
        "bucket": "ai2-s2-research",
        "path": "scibert/huggingface_pytorch/scibert_scivocab_uncased.tar",
        "file_name": "scibert_scivocab_uncased.tar"
    },
    'scibert_scivocab_cased': {
        "bucket": "ai2-s2-research",
        "path": "scibert/huggingface_pytorch/scibert_scivocab_cased.tar",
        "file_name": "scibert_scivocab_cased.tar"
    }
}



from sklearn.model_selection import train_test_split
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.metrics import f1_score
import tensorflow as tf

from wellcomeml.ml.attention import HierarchicalAttention
from wellcomeml.ml.keras_utils import Metrics

class BiLSTMClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, learning_rate=0.01, batch_size=32, nb_epochs=5,
                 dropout=0.1, nb_layers=2, multilabel=False,
                 attention=False):
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.nb_epochs = nb_epochs
        self.dropout = dropout
        self.nb_layers = nb_layers
        self.multilabel = multilabel
        self.attention = attention

    def fit(self, X, Y, embedding_matrix=None, *_):
        sequence_length = X.shape[1]
        vocab_size = X.max() + 1
        embedding_size = embedding_matrix.shape[1] if embedding_matrix else 100
                                 
        nb_outputs = Y.max() if not self.multilabel else Y.shape[1]
        output_activation = 'sigmoid' if nb_outputs==1 or self.multilabel else 'softmax'

        X_train, X_val, Y_train, Y_val = train_test_split(X, Y, test_size=0.1, shuffle=True)

        def residual_bilstm(x1, l2):
            x2 = tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(int(x1.shape[-1]/2), return_sequences=True, kernel_regularizer=l2))(x1)
            return tf.keras.layers.add([x1, x2])

        def residual_attention(x1):
            x2 = HierarchicalAttention()(x1)
            x2 = tf.keras.layers.Dropout(self.dropout)(x2)
            x2 = tf.keras.layers.LayerNormalization()(x2)
            return tf.keras.layers.add([x1, x2])

        l2 = tf.keras.regularizers.l2(1e-6)
        embeddings_initializer = tf.keras.initializers.Constant(embedding_matrix) if embedding_matrix else 'uniform'
        inp = tf.keras.layers.Input(shape=(sequence_length,))
        x = tf.keras.layers.Embedding(
                vocab_size,
                embedding_size,
                input_length=sequence_length,
                embeddings_initializer=embeddings_initializer
            )(inp)
        for _ in range(self.nb_layers):
            x = residual_bilstm(x, l2)
        if self.attention:
            x = residual_attention(x)
        x = tf.keras.layers.GlobalMaxPooling1D()(x)
        x = tf.keras.layers.Dense(20, kernel_regularizer=l2)(x)
        out = tf.keras.layers.Dense(nb_outputs, activation=output_activation, kernel_regularizer=l2)(x)
        self.model = tf.keras.Model(inp, out)

        optimizer = tf.keras.optimizers.Adam(lr=self.learning_rate, clipnorm=1.0)
        metrics = Metrics(validation_data=(X_val, Y_val))
        self.model.compile(optimizer=optimizer, loss='binary_crossentropy', metrics=[])
        self.model.fit(X_train, Y_train, epochs=self.nb_epochs, batch_size=self.batch_size, validation_data=(X_val, Y_val), callbacks=[metrics])
        return self

    def predict(self, X, *_):
        return self.model(X).numpy() > 0.5

    def score(self, X, Y):
        Y_pred = self.predict(X)
        return f1_score(Y, Y_pred, average='micro')



from collections import defaultdict
import os

import tensorflow as tf
from transformers import (
    BertConfig,
    BertTokenizer,
    TFBertForSequenceClassification
)
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import train_test_split
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError


class SemanticEquivalenceClassifier(BaseEstimator, TransformerMixin):
    """
    Class to fine-tune BERT-type models for semantic equivalence, for example
    paraphrase, textual similarity and other NLU tasks
    """
    def __init__(self, pretrained='bert',
                 batch_size=32,
                 eval_batch_size=32*2,
                 learning_rate=3e-5,
                 test_size=0.2,
                 max_length=128):
        self.pretrained = pretrained
        self.batch_size = batch_size
        self.eval_batch_size = eval_batch_size
        self.learning_rate = learning_rate
        self.test_size = test_size
        self.max_length = max_length

        # Defines attributes that will be initialised later
        self.config = None
        self.tokenizer = None
        self.model = None
        self.train_dataset = None
        self.valid_dataset = None
        self.train_steps = None
        self.valid_steps = None

        self.history = defaultdict(list)

    # Bert models have a tensorflow checkopoint, otherwise,
    # we need to load the pytorch versions with the parameter `from_pt=True`

    def _initialise_models(self):
        if self.pretrained == 'bert':
            model_name = 'bert-base-cased'
            from_pt = False
        elif self.pretrained == 'scibert':
            model_name = 'allenai/scibert_scivocab_cased'
            from_pt = True

        self.config = BertConfig.from_pretrained(model_name, num_labels=2)
        self.tokenizer = BertTokenizer.from_pretrained(model_name)
        self.model = TFBertForSequenceClassification.from_pretrained(
            model_name, config=self.config, from_pt=from_pt
        )

    def _prep_dataset_generator(self, X, y):
        batch_encoding = self.tokenizer.batch_encode_plus(
            X, max_length=self.max_length, add_special_tokens=True,
        )

        def gen_train():
            for i in range(len(X)):
                features = {k: pad(batch_encoding[k][i], self.max_length)
                            for k in batch_encoding}

                yield (features, int(y[i]))

        dataset = tf.data.Dataset.from_generator(
            gen_train,
            ({"input_ids": tf.int32,
              "attention_mask": tf.int32,
              "token_type_ids": tf.int32}, tf.int64),
            (
                {
                    "input_ids": tf.TensorShape([None]),
                    "attention_mask": tf.TensorShape([None]),
                    "token_type_ids": tf.TensorShape([None]),
                },
                tf.TensorShape([]),
            ),
        )

        return dataset

    def fit(self, X, y, random_state=None, epochs=3):
        """
        Fits a sentence similarity model

        Args:
            X: List of 2-uples: [('Sentence 1', 'Sentence 2'), ....]
            y: list of 0 (not similar) and 1s (similar)
            random_state: a random state for the train_test_split
            epochs: number_of epochs

        Returns:

        """
        # Initialises/downloads model if not trained before.
        # If trained, fits extra epochs without the transformations
        try:
            check_is_fitted(self)
        except NotFittedError:
            self._initialise_models()
            
            # Train/val split
            X_train, X_valid, y_train, y_valid = train_test_split(
                X, y, test_size=self.test_size, random_state=random_state
            )
    
            # Generates tensorflow dataset
            self.train_dataset = self._prep_dataset_generator(X_train, y_train)
            self.valid_dataset = self._prep_dataset_generator(X_valid, y_valid)
    
            # Generates mini-batches and stores in a class variable
            self.train_dataset = self.train_dataset.shuffle(self.max_length).\
                batch(self.batch_size).repeat(-1)
            self.valid_dataset = self.valid_dataset.batch(self.eval_batch_size)
    
            opt = tf.keras.optimizers.Adam(learning_rate=self.learning_rate,
                                           epsilon=1e-08)
    
            loss = tf.keras.losses.SparseCategoricalCrossentropy(
                from_logits=True
            )
    
            # Train and evaluate using tf.keras.Model.fit()
    
            metric = tf.keras.metrics.SparseCategoricalAccuracy("accuracy")
    
            self.model.compile(optimizer=opt, loss=loss, metrics=[metric])
    
            self.train_steps = len(X_train) // self.batch_size
            self.valid_steps = len(X_valid) // self.eval_batch_size
        finally:
            history = self.model.fit(
                self.train_dataset,
                epochs=epochs,
                steps_per_epoch=self.train_steps,
                validation_data=self.valid_dataset,
                validation_steps=self.valid_steps
            )

            # Accumulates history of metrics from different partial fits
            for metric, values in history.history.items():
                self.history[metric] += values

            self.trained_ = True

        return self

    def score(self, X):
        """
        Calculates scores for model prediction

        Args:
            X: List of 2-uples: [('Sentence 1', 'Sentence 2'), ....]

        Returns:
            An array of shape len(X) x 2 with scores for classes 0 and 1

        """
        X_tokenized = self.tokenizer.batch_encode_plus(
            X, max_length=self.max_length, add_special_tokens=True,
            pad_to_max_length=True,
            return_tensors="tf"
        )
        predictions = tf.convert_to_tensor(self.model.predict(X_tokenized))

        return tf.keras.activations.softmax(predictions).numpy()

    def predict(self, X):
        """
        Calculates the predicted class (0 - for negative and 1 for positive)
        for X.

        Args:
            X: List of 2-uples: [('Sentence 1', 'Sentence 2'), ....]

        Returns:
            An array of 0s and 1s

        """
        return self.score(X).argmax(axis=1)

    def save(self, path):
        os.makedirs(path, exist_ok=True)
        self.model.save_pretrained(path)


def pad(x, pad_len):
    """Old versions of transformers do not pad by default"""
    return x + [0]*(pad_len-len(x))



import os
import tarfile

import boto3
from botocore import UNSIGNED
from botocore.client import Config

from wellcomeml.logger import logger

DATA_DIR = os.path.expanduser("~/.cache/wellcomeml/data")
DATA_DISPATCH = {
    "hoc": {
        "bucket": "datalabs-public",
        "path": "datasets/hoc/hoc.tar",
        "file_name": "hoc.tar"
    }
}

def check_cache_and_download(dataset_name):
    """ Checks if dataset_name is cached and return complete path"""
    os.makedirs(DATA_DIR, exist_ok=True)

    dataset_path = os.path.join(DATA_DIR, dataset_name)
    if not os.path.exists(dataset_path):
        logger.info(f"Could not find dataset {dataset_name}. Downloading from S3")

        # The following allows to download from S3 without AWS credentials
        s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        tmp_file = os.path.join(
            DATA_DIR,
            DATA_DISPATCH[dataset_name]['file_name']
        )

        s3.download_file(
            DATA_DISPATCH[dataset_name]['bucket'],
            DATA_DISPATCH[dataset_name]['path'], tmp_file
        )

        tar = tarfile.open(tmp_file)
        tar.extractall(path=DATA_DIR)
        tar.close()

        os.remove(tmp_file)

    return dataset_path



__name__='wellcomeml'
__version__='2020.5.0'
__description__='Utilities for managing nlp models and for processing text-related data at Wellcome Data Labs'
__url__='https://github.com/wellcometrust/datalabs/tree/master/code/wellcomeml/'
__author__='Wellcome Trust DataLabs Team'
__author_email__='Grp_datalabs-datascience@Wellcomecloud.onmicrosoft.com'
__license__='MIT'



#!/usr/bin/env python3
# coding: utf-8

"""
Utilities for loading and saving data from various formats
"""

import json

from wellcomeml.logger import logger


def write_jsonl(input_data, output_file):
    """
    Write a dict to jsonl (line delimited json)

    Output format will look like:

    ```
    {'a': 0}
    {'b': 1}
    {'c': 2}
    {'d': 3}
    ```

    Args:
        input_data(dict): A dict to be written to json.
        output_file(str): Filename to which the jsonl will be saved.
    """

    with open(output_file, 'w') as fb:

        # Check if a dict (and convert to list if so)

        if isinstance(input_data, dict):
            input_data = [value for key, value in input_data.items()]

        # Write out to jsonl file

        logger.debug('Writing %s lines to %s', len(input_data), output_file)

        for i in input_data:
            json_ = json.dumps(i) + '\n'
            fb.write(json_)


def _yield_jsonl(file_name):
    for row in open(file_name, "r"):
        yield json.loads(row)


def read_jsonl(input_file):
    """Create a list from a jsonl file

    Args:
        input_file(str): File to be loaded.
    """

    out = list(_yield_jsonl(input_file))

    logger.debug('Read %s lines from %s', len(out), input_file)

    return out



"""
A class that for each of a list of sentences will find the most similar document in a corpus
using the TFIDF vectors or a BERT embedding from the corpus documents.

"""

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import f1_score
from sklearn.metrics.pairwise import cosine_similarity
import numpy as np

from wellcomeml.ml import BertVectorizer


class SimilarityEntityLinker():
    def __init__(self, stopwords, embedding='tf-idf'):
        """
        Input: 
            stopwords - list of stopwords
            embedding - How to embed the documents
                    in order to find which document in the corpus
                    is most similar to the sentence.
                    embedding='tf-idf': Use a TFIDF vectoriser
                    embedding='bert': Use a BERT vectoriser
        """

        self.stopwords = stopwords
        self.embedding = embedding

    def _clean_text(self, text):
        """
        Clean a body of text
        """
        text_split = text.replace("\n"," ")

        return text_split

    def _clean_kb(self, raw_knowledge_base):
        """
        Creates a cleaned version of the raw_knowledge_base
        which is a dictionary of each document's text.
        
        Don't include any empty text information
        """

        knowledge_base = {}
        for key, text in raw_knowledge_base.items():
            if len(text.replace(" ","")) != 0:
                knowledge_base[key] = self._clean_text(text)

        return knowledge_base

    def fit(self, documents):
        """
        documents: dictionary of the texts from each of the corpus documents
        """

        documents = self._clean_kb(documents)

        document_texts = list(documents.values())
        self.classifications = list(documents.keys())

        if self.embedding == 'tf-idf':
            self.vectorizer = TfidfVectorizer(stop_words=self.stopwords)
            self.corpus_matrix = self.vectorizer.fit_transform(document_texts)
        else:
            self.vectorizer = BertVectorizer(sentence_embedding='mean_last')
            self.vectorizer.fit()
            self.corpus_matrix = self.vectorizer.transform(document_texts)

    def predict_proba(self, data):
        """
        Returns probability estimates for each class
        in the same order as self.classifications
        """

        sentences = [self._clean_text(sentence) for sentence, _ in data]
        query = self.vectorizer.transform(sentences)

        class_probabilities = cosine_similarity(query, self.corpus_matrix)

        return class_probabilities

    def optimise_threshold(self, data, id_col='orcid', no_id_col='No ORCID'):
        """
        Find the f1 scores when different similarity thresholds are used.
        Use half the maximum F1 score found as the optimal threshold.
        This value will be used as the value for the similarity
        threshold in predict, unless another value is given.
        """

        y_true = [test_meta[id_col] for _, test_meta in data]
        
        # Find the best prediction and probability for each data point 
        probabilities = self.predict_proba(data)
        pred_entities = []
        pred_similarities = []
        for entity, similarity in zip(
                np.argmax(probabilities, axis=1),
                np.max(probabilities, axis=1)
                ):
            pred_entities.append(self.classifications[entity])
            pred_similarities.append(similarity)

        # Lipton, Z. C., Elkan, C., & Naryanaswamy, B. (2014)
        # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4442797/
        f1_scores = []
        for similarity_threshold in np.linspace(0, 1, 40):
            pred_entities_temp = [
                pred_entities[i] if sim <= similarity_threshold else no_id_col for i, sim in enumerate(pred_similarities)
                ]
            f1_scores.append(
                f1_score(y_true, pred_entities_temp, average='weighted', zero_division=0)
                )
        self.optimal_threshold = max(f1_scores)/2

    def predict(self, data, similarity_threshold=None, no_id_col='No ORCID'):
        """
        Identify the most similar document to a sentence using TFIDF

        If the most similar document doesnt have a similarity value over
        a threshold then return they key 'No ORCID'

        similarity_threshold can be specified, otherwise if you've optimised
        the threshold it will use this value

        Input:
            data: a list of tuples in the form
                [('A sentence about Farrar',
                {metadata}),
                ('A sentence about Smith',
                {metadata})]
                For this predict function the contents of 
                {metadata} isn't important
            similarity_threshold: The threshold by which to 
                classify a match as being true or that there is
                no match. If this is None then the best threshold 
                will be found
        Output:
            pred_entities: a list of predictions
                of which document in the corpus each data
                point is likely to link to
                ['0000-0002-2700-623X', '0000-0002-6259-1606', 'No ORCID']
        """

        if similarity_threshold is None:
            similarity_threshold = self.optimal_threshold

        # Find all the probabilities for the different classes
        probabilities = self.predict_proba(data)

        pred_entities = []
        for entity, similarity in zip(
                np.argmax(probabilities, axis=1),
                np.max(probabilities, axis=1)
                ):
            if similarity > similarity_threshold:
                pred_entities.append(self.classifications[entity])
            else:
                pred_entities.append(no_id_col)

        return pred_entities

    def evaluate(self, y_true, y_pred):

        f1_micro = f1_score(y_true, y_pred, average = 'micro', zero_division = 0)

        return f1_micro



import sys

from wellcomeml.viz.colors import WellcomeColor,\
    NAMED_COLORS_DICT, NAMED_COLORS_LARGE_DICT
from typing import List
from types import ModuleType

# The palette module uses a trick to prevent users using the palettes from
# accidentally changing it, by defining them as a class property.


class _WellcomePalette(ModuleType):
    """Represents a Wellcome palette"""

    __all__: List[str] = []

    @property
    def Wellcome11(self) -> List[WellcomeColor]:
        """
        Full Wellcome categorical palette as in

        https://company-57536.frontify.com/d/gFEfjydViLRJ/
        wellcome-brand-book#/visuals/dataviz-elements-and-rationale

        Returns:
            List of Wellcome colors

        """
        return list(NAMED_COLORS_DICT.values())

    @property
    def Wellcome33Shades(self) -> List[WellcomeColor]:
        """Wellcome33 with 30% decrements palette"""
        return list(NAMED_COLORS_LARGE_DICT.values())

    @property
    def WellcomeMatrix(self) -> List[List[WellcomeColor]]:
        """Matrix palette"""
        return [
            [self.Wellcome33Shades[i],
             self.Wellcome33Shades[i + 1],
             self.Wellcome33Shades[i + 2]]
            for i in range(0, len(self.Wellcome33Shades), 3)
        ]

    @property
    def Wellcome33(self) -> List[WellcomeColor]:
        """Linearised matrix palette with no repeated adjacent color"""
        return [color for color, _, _ in self.WellcomeMatrix] + \
               [color for _, color, _ in self.WellcomeMatrix] + \
               [color for _, _, color in self.WellcomeMatrix]

    @property
    def WellcomeBackground(self) -> WellcomeColor:
        """Wellcome background color"""
        return WellcomeColor("Backgrounds", "#F2F4F6", (244, 244, 246))

    @property
    def WellcomeNoData(self) -> WellcomeColor:
        """Wellcome color for 'noise' or missing data or 'other' category """
        return WellcomeColor("No data", "#CCD8DD", (204, 216, 221))


# Transfers the class property to module level variables
_mod = _WellcomePalette('wellcomeml.viz.palettes')
_mod.__doc__ = __doc__
_mod.__all__ = dir(_mod)
sys.modules['wellcomeml.viz.palettes'] = _mod

del _mod, sys



"""
TODO: Fill this
"""
from pathlib import Path
import random
import os

from spacy.vocab import Vocab
from spacy.kb import KnowledgeBase
from spacy.symbols import PERSON
from spacy.tokens import Span
from spacy.util import minibatch, compounding
import spacy

from wellcomeml.ml.spacy_knowledge_base import SpacyKnowledgeBase


class SpacyEntityLinker(object):
    def __init__(self, kb_path, n_iter=50, print_output=True):
        self.kb_path = kb_path
        self.n_iter = n_iter
        self.print_output = print_output

    def _remove_examples_not_in_kb(self, kb, data):
        # Remove examples with unknown identifiers to the knowlege base
        kb_ids = kb.get_entity_strings()
        for text, annotation in data:
            for offset, kb_id_dict in annotation["links"].items():
                new_dict = {}
                for kb_id, value in kb_id_dict.items():
                    if kb_id in kb_ids:
                        new_dict[kb_id] = value
                    else:
                        print(
                            "Removed", kb_id,
                            "from training because it is not in the KB."
                        )
                annotation["links"][offset] = new_dict
        return data

    def train(self, data):
        """
        Input:
            data: list of training data in the form
                [('A sentence about Farrar',
                {'links': {(17, 22): {'Q1': 1.0, 'Q2': 0.0}}})]

        See https://spacy.io/usage/linguistic-features#entity-linking for where I got this code from
        """
        # TODO: Replace n_iter with self.n_iter
        n_iter = self.n_iter

        kb = SpacyKnowledgeBase()
        kb = kb.load(self.kb_path)
        print("Loaded Knowledge Base from '%s'" % self.kb_path)

        nlp = spacy.blank("en", vocab=kb.vocab)
        nlp.vocab.vectors.name = "spacy_pretrained_vectors"
    
        entity_linker = nlp.create_pipe("entity_linker")
        entity_linker.set_kb(kb)
        nlp.add_pipe(entity_linker, last=True)        

        data = self._remove_examples_not_in_kb(kb, data)
        
        other_pipes = [pipe for pipe in nlp.pipe_names if pipe != "entity_linker"]
        with nlp.disable_pipes(*other_pipes):
            optimizer = nlp.begin_training()
            for itn in range(n_iter):
                random.shuffle(data)
                losses = {}
                batches = minibatch(data, size=compounding(4.0, 32.0, 1.001))
                for batch in batches:
                    texts, annotations = zip(*batch)
                    nlp.update(
                        texts,
                        annotations,
                        drop=0.2,
                        losses=losses,
                        sgd=optimizer,
                    )
                if self.print_output:
                    print(itn, "Losses", losses)

        self.nlp = nlp
        return self.nlp

    def _get_token_nums(self, doc, char_idx):
        """
        Convert a character index to a token index
        i.e. what number token is character number char_idx in ?
        """
        for i, token in enumerate(doc):
            if char_idx > token.idx:
                continue
            if char_idx == token.idx:
                return i
            if char_idx < token.idx:
                return i

    def predict(self, data):
        """
        See how well the model predicts which entity you are referring to in your data
        Input:
            data: list of test data in the form
                [('A sentence about Farrar',
                {'links': {(17, 22): {'Q1': 1.0, 'Q2': 0.0}}})]
        Output:
           pred_entities_ids: ['Q1', 'Q1', 'Q2']
    """
        # TODO: Replace nlp_el with self.nlp 
        nlp_el = self.nlp

        # IMPORTANT
        #
        # This code assumes each training example has one entity
        # e.g. (start, end) = list(values['links'].keys())[0]
        #
        # If the code has more than one entities we predict the
        # first only  e.g. pred_entity_id = doc.ents[0].kb_id_
        
        pred_entities_ids = []
        for text, values in data:
            (start, end) = list(values['links'].keys())[0]
            entity_text = text[start:end]

            doc = nlp_el.tokenizer(text)
            entity_tokens = nlp_el.tokenizer(entity_text)

            # Set entity span to PERSON
            doc.ents = [
                Span(
                    doc,
                    self._get_token_nums(doc, start),
                    self._get_token_nums(doc, end),
                    label=PERSON
                )
            ]

            doc = nlp_el.get_pipe("entity_linker")(doc)

            pred_entity_id = doc.ents[0].kb_id_
            pred_entities_ids.append(pred_entity_id)

        return pred_entities_ids

    def save(self, output_dir):
        output_dir = Path(output_dir)
        if not output_dir.exists():
            output_dir.mkdir()
        self.nlp.to_disk(output_dir)
        print("Saved model to", output_dir)

    def load(self, output_dir):
        print("Loading from", output_dir)
        self.nlp = spacy.load(output_dir)
        return self.nlp



#!/usr/bin/env python3
# coding: utf-8

"""
A generic "frequency" vectorizer that wraps all usual transformations.
"""
import re

import numpy as np
import spacy
from sklearn.feature_extraction.text import TfidfVectorizer

from wellcomeml.logger import logger
nlp = spacy.load('en_core_web_sm', disable=['ner', 'tagger', 'parser',
                                            'textcat'])


class WellcomeTfidf(TfidfVectorizer):
    """
    Class to wrap some basic transformation and text
    vectorisation/embedding
    """
    def __init__(self, **kwargs):
        """

        Args:
            Any sklearn "tfidfvectorizer" arguments (min_df, etc.)

        """
        self.embedding = 'tf-idf'

        logger.info("Initialising frequency vectorizer.")

        kwargs['stop_words'] = kwargs.get('stop_words', 'english')

        super().__init__(**kwargs)

    @classmethod
    def save_transformed(cls, path, X_transformed):
        """Saves transformed embedded vectors"""
        # Numpy will save the transformed sparse array
        # into a 0-dimensional array, preserving sparsity
        np.save(path, X_transformed)

    @classmethod
    def load_transformed(cls, path):
        """Loads transformed embedded vectors"""

        # Loads the 0-dimensional array and returns its content
        X_transformed = np.load(path, allow_pickle=True)

        return X_transformed[()]

    def regex_transform(self, X, remove_numbers='years', *_):
        """
        Extra regular expression transformations to clean text
        Args:
            X: A list of texts (strings)
            *_:
            remove_numbers: Whether to remove years or all digits. Caveat:
            This does not only remove years, but **any number** between
            1000 and 2999.

        Returns:
            A list of texts with the applied regex transformation

        """
        if remove_numbers == 'years':
            return [re.sub(r'[1-2]\d{3}', '', text) for text in X]
        elif remove_numbers == 'digits':
            return [re.sub(r'\d', '', text) for text in X]
        else:
            return X

    def spacy_lemmatizer(self, X, remove_stopwords_and_punct=True):
        """
        Uses spacy pre-trained lemmatisation model to
        Args:
            X: A list of texts (strings)
            remove_stopwords_and_punct: Whether to remove stopwords,
            punctuation, pronouns

        Returns:

        """
        logger.info("Using spacy pre-trained lemmatiser.")
        if remove_stopwords_and_punct:
            return [
                [token.lemma_.lower() for token in doc
                 if not token.is_stop and not token.is_punct and
                 token.lemma_ != "-PRON-"]
                for doc in nlp.tokenizer.pipe(X)
            ]
        else:
            return [[token.lemma_.lower() for token in
                     doc] for doc in nlp.tokenizer.pipe(X)]

    def transform(self, X, regex=True, spacy_lemmatizer=True, *_):
        if regex:
            X = self.regex_transform(X)
        if spacy_lemmatizer:
            X = self.spacy_lemmatizer(X)

        X = [' '.join(text) for text in X]

        return super().transform(X)

    def fit(self, X, regex=True, spacy_lemmatizer=True, *_):
        if regex:
            X = self.regex_transform(X)
        if spacy_lemmatizer:
            X = self.spacy_lemmatizer(X)

        logger.info("Fitting vectorizer.")

        X = [' '.join(text) for text in X]
        super().fit(X)

        return self



#!/usr/bin/env python
# coding: utf8
"""
Train a convolutional neural network for multilabel classification
of grants
Adapted from https://github.com/explosion/spaCy/blob/master/examples/training/train_textcat.py
"""
from spacy.util import minibatch, compounding, decaying
from sklearn.metrics import f1_score, precision_score, recall_score
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split
from sklearn.metrics import precision_recall_fscore_support
from scipy.sparse import csr_matrix
import numpy as np
import spacy
import torch

import random
import time

from wellcomeml.logger import logger

is_using_gpu = spacy.prefer_gpu()
if is_using_gpu:
    torch.set_default_tensor_type("torch.cuda.FloatTensor")


class SpacyClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, threshold=0.5, n_iterations=5,
                 batch_size=8, learning_rate=0.001,
                 dropout=0.1, shuffle=True, architecture="simple_cnn",
                 multilabel=True, pre_trained_vectors_path=None):
        """
        Args:
            threshold: the threshold above of which a label should be assigned.
                Should take values from 0 to 1. default is 0.5.
            batch_size: the number of examples that will be given to the optimizer
                in each update of the parameters. default is 8.
            dropout: the dropout added to the layers. default is 0.1.
            learning_rate: the learning rate of the optimizer. default is 0.001.
            n_iterations: number of iterations on the entire dataset (epochs)
                default is 5
            shuffle: whether the example are shuffled before an iterations.
                default is True
            architecture: architecture that Spacy uses. can be one of "bow", "simple_cnn"
                and "ensemble". default is "simple_cnn"
            multilabel: whether the problem is multilabel and thus Y is a matrix. default is True
            pre_trained_vectors_path: path to pretrained vectors produced by spacy pretrain.
                default is None
        """
        self.threshold = threshold
        self.batch_size = batch_size
        self.dropout = dropout
        self.learning_rate = learning_rate
        self.n_iterations=n_iterations
        self.shuffle=shuffle
        self.architecture=architecture
        self.pre_trained_vectors_path=pre_trained_vectors_path
        self.multilabel=multilabel

    def _init_nlp(self):
        self.nlp = spacy.blank('en')

        self.textcat = self.nlp.create_pipe(
            "textcat",
            config={
                "architecture": self.architecture,
                "exclusive_classes": not self.multilabel
            }
        )
        self.nlp.add_pipe(self.textcat, last=True)

        for label in self.unique_labels:
            self.textcat.add_label(label)

    def load(self, model_dir):
        self.nlp = spacy.load(model_dir)
        # another hack to get the labels from textcat since
        # spacy does not serialise every attribute
        for pipe_name, pipe in self.nlp.pipeline:
            # to cover case of textcat and trf_textcat
            if 'textcat' in pipe_name:
                self.unique_labels = pipe.labels

    def save(self, output_dir):
        self.nlp.to_disk(output_dir)

    def _label_binarizer_inverse_transform(self, Y_train):
        "Transforms Y matrix to labels which are the non zero indices"
        data = []
        for row in Y_train:
            if type(row) is csr_matrix:
                row = row.todense()
                row = np.squeeze(np.asarray(row))
            row_data = [str(i) for i, item in enumerate(row) if item]
            data.append(row_data)
        # Do we need to convert to numpy?
        return np.array(data)

    def fit(self, X, Y):
        """
        Args:
            X: 1d list of documents
            Y: 2d numpy array (nb_examples, nb_labels)
        TODO: Generalise to y being 1d
        """
        if type(Y) in [list, tuple]:
            Y = np.array(Y)

        X_train, X_test, Y_train, Y_test = train_test_split(
            X, Y, random_state=42
        )
        # Free memory
        del X
        del Y

        if self.multilabel:
            nb_labels = Y_train.shape[1]
            self.unique_labels = [str(i) for i in range(nb_labels)]
        else:
            self.unique_labels = np.unique(Y_train)
        self._init_nlp()
        n_iter = self.n_iterations
        
        def yield_train_data(X_train, Y_train):
            for x, y in zip(X_train, Y_train):
                if self.multilabel:
                    tags = self._label_binarizer_inverse_transform([y])[0]
                    cats = {
                        label: label in tags
                        for label in self.unique_labels
                    }
                else:
                    cats = {
                        label: label == y
                        for label in self.unique_labels
                    }
                yield (x, {"cats": cats})
        
        other_pipes = [pipe for pipe in self.nlp.pipe_names if pipe != "textcat"]
        with self.nlp.disable_pipes(*other_pipes):  # only train textcat
            optimizer = self.nlp.begin_training()
            optimizer.alpha = self.learning_rate
            #optimizer.L2 = 1e-4

            if self.pre_trained_vectors_path:
                with open(self.pre_trained_vectors_path, "rb") as f:
                    self.textcat.model.tok2vec.from_bytes(f.read())
            
            logger.info("Training the model...")
            logger.info("{:^5}\t{:^5}\t{:^5}\t{:^5}\t{:^5}\t{:^5}".format("ITER", "LOSS", "P", "R", "F", "SPEED"))
            batch_sizes = compounding(4.0, self.batch_size, 1.001)
            #dropout = decaying(0.6, 0.2, 1e-4)
            for i in range(n_iter):
                start_time = time.time()
                nb_examples = 0
                losses = {}

                def shuffle(X_train, Y_train):
                    # this is not memory friendly but there should be
                    # memory from deleting X, Y
                    d = list(zip(X_train, Y_train))
                    random.shuffle(d)
                    X_train, Y_train = zip(*d)
                    return X_train, Y_train

                if self.shuffle:
                    X_train, Y_train = shuffle(X_train, Y_train)

                train_data = yield_train_data(X_train, Y_train)
                batches = minibatch(train_data, size=batch_sizes)
                for batch in batches:
                    texts, annotations = zip(*batch)
                    next_dropout = self.dropout # next(dropout)
                    self.nlp.update(texts, annotations, sgd=optimizer, drop=next_dropout, losses=losses)
                    nb_examples += len(texts)
                end_time = time.time() - start_time
                examples_per_second = round(nb_examples / end_time, 2)
                with self.textcat.model.use_params(optimizer.averages):
                    Y_train_pred = self.predict(X_train)
                    Y_test_pred = self.predict(X_test)
                    p, r, f, _ = precision_recall_fscore_support(Y_test, Y_test_pred, average='micro')
                loss = losses["textcat"]
                logger.info(
                    "{0:2d}\t\t{1:.3f}\t{2:.3f}\t{3:.3f}\t{4:.3f}\t{5:.2f} ex/s".format(
                        i,
                        loss,
                        p,
                        r,
                        f,
                        examples_per_second
                    )
                )
        return self

    def partial_fit(self, X, Y, classes=None):
        """
        Args:
            X: 1d list of documents
            Y: 2d numpy array (nb_examples, nb_labels)
        TODO: Generalise to y being 1d
        """
        Y = np.array(Y)

        if not hasattr(self, 'unique_labels'):
            # We could also use classes here
            nb_labels = Y.shape[1]
            self.unique_labels = [str(i) for i in range(nb_labels)]
            self._init_nlp()

        texts = X

        if self.multilabel:
            train_tags = self._label_binarizer_inverse_transform(Y)
            train_cats = [
                {
                    label: label in tags
                    for label in self.unique_labels
                } for tags in train_tags
            ]
        else:
            train_cats = []
            for y in Y:
                cats = {
                    label: label == y
                    for label in self.unique_labels
                }
                train_cats.append(cats)
        annotations = [{"cats": cats} for cats in train_cats]

        other_pipes = [pipe for pipe in self.nlp.pipe_names if pipe != "textcat"]
        with self.nlp.disable_pipes(*other_pipes):  # only train textcat
            if not hasattr(self, 'optimizer'):
                self.optimizer = self.nlp.begin_training()
                self.optimizer.alpha = self.learning_rate

            self.nlp.update(texts, annotations, sgd=self.optimizer, drop=self.dropout)

        return self
    
    def predict(self, X):
        def transform_output(doc):
            cats = doc.cats
            if self.multilabel:
                out = [
                    1 if cats[label] > self.threshold else 0
                    for label in self.unique_labels
                ]
            else:
                out = max(cats.items(), key=lambda x: x[1])[0]
            return out
        docs = self.nlp.pipe(X)
        return np.array([transform_output(doc) for doc in docs])

    def predict_proba(self, X):
        def get_proba(x):
            cats = self.nlp(x).cats
            out = [
                cats[label]
                for label in self.unique_labels
            ]
            return out
        return np.array([get_proba(x) for x in X])

    def score(self, X, Y):
        Y = np.array(Y)
        Y_pred = self.predict(X)
        return f1_score(Y_pred, Y, average='micro')



import os
from wellcomeml.logger import logger

# Introduced a development_transformers env variable, that allows to
# disable functions that use spacy.

development_transformers_mode = \
    os.environ.get("WELLCOMEML_ENV", "") == "development_transformers"

if development_transformers_mode:
    logger.warning("Running in development mode. Only loading modules that"
                   " use new version of transformers.")

    from .bert_semantic_equivalence import SemanticEquivalenceClassifier
else:
    from .frequency_vectorizer import WellcomeTfidf

    try:
        from .vectorizer import Vectorizer
        from .spacy_ner import SpacyNER
        from .spacy_classifier import SpacyClassifier
        from .bert_classifier import BertClassifier
        from .bert_vectorizer import BertVectorizer
        from .spacy_knowledge_base import SpacyKnowledgeBase
        from .spacy_entity_linking import SpacyEntityLinker
        from .similarity_entity_linking import SimilarityEntityLinker
        from .bert_semantic_equivalence import SemanticEquivalenceClassifier
        from .cnn import CNNClassifier
        from .bilstm import BiLSTMClassifier
        from .keras_vectorizer import KerasVectorizer
    except ImportError:
        logger.warning("Using WellcomeML without extras (transformers & torch).")



"""
"""
from sklearn.base import BaseEstimator, TransformerMixin
from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.preprocessing.sequence import pad_sequences

class KerasVectorizer(BaseEstimator, TransformerMixin):
    def __init__(self, vocab_size=None, sequence_length=None, oov_token='<OOV>'):
        self.vocab_size = vocab_size
        self.oov_token = oov_token
        self.sequence_length = sequence_length

    def fit(self, X, *_):
        self.tokenizer = Tokenizer(
            num_words=self.vocab_size,
            oov_token=self.oov_token
        )
        self.tokenizer.fit_on_texts(X)
        return self

    def transform(self, X, *_):
        sequences = self.tokenizer.texts_to_sequences(X)
        return pad_sequences(sequences, maxlen=self.sequence_length)

    def build_embedding_matrix(embeddings_path):
        embeddings_index = {}
        with open(embeddings_path) as f:
            for line in f:
                word, coefs = line.split(maxsplit=1)
                coefs = np.fromstring(coefs, 'f', sep=' ')
                embeddings_index[word] = coefs

        emb_dim = len(coefs)
        num_words = len(self.tokenizer.word_index) + 1

        embedding_matrix = np.zeros((num_words, emb_dim))
        for word, i in word_index.items():
            embedding_vector = embeddings_index.get(word)
            if embedding_vector is not None:
                # words not found in embedding index will be all-zeros.
                embedding_matrix[i] = embedding_vector
        return embedding_matrix



"""
CNN architecture inspired from spacy for NLP tasks

It follows the embed, encode, attend, predict framework
Embed: learns a new embedding but can receive
    pre trained embeddings as well
Encode: stacked CNN with context window 3 that maintains
    the size of input and applies dropout and layer norm
Attend: not yet implemented
Predict: softmax or sigmoid depending on number of outputs
    and whether task is multilabel
"""
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score, precision_score, recall_score
import tensorflow as tf

from wellcomeml.ml.attention import HierarchicalAttention
from wellcomeml.ml.keras_utils import Metrics

class CNNClassifier(BaseEstimator, ClassifierMixin):
    def __init__(self, context_window = 3, learning_rate=0.001,
                 batch_size=32, nb_epochs=5, dropout=0.2,
                 nb_layers=4, hidden_size=100, multilabel=False,
                 attention=False):
        self.context_window = context_window
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.nb_epochs = nb_epochs
        self.dropout = dropout
        self.nb_layers = nb_layers
        self.hidden_size = hidden_size # note that on current implementation CNN use same hidden size as embedding so if embedding matrix is passed, this is not used. in the future we can decouple
        self.multilabel = multilabel
        self.attention = attention

    def fit(self, X, Y, embedding_matrix=None):
        sequence_length = X.shape[1]
        vocab_size = X.max() + 1
        emb_dim = embedding_matrix.shape[1] if embedding_matrix else self.hidden_size
        nb_outputs = Y.max() if not self.multilabel else Y.shape[1]

        def residual_conv_block(x1):
            filters = x1.shape[2]
            x2 = tf.keras.layers.Conv1D(
                filters,
                self.context_window,
                padding='same',
                activation='relu',
                kernel_regularizer=tf.keras.regularizers.l2(1e-6)
            )(x1)
            x2 = tf.keras.layers.Dropout(self.dropout)(x2)
            x2 = tf.keras.layers.LayerNormalization()(x2)
            return tf.keras.layers.add([x1, x2])

        def residual_attention(x1):
            x2 = HierarchicalAttention()(x1)
            x2 = tf.keras.layers.Dropout(self.dropout)(x2)
            x2 = tf.keras.layers.LayerNormalization()(x2)
            return tf.keras.layers.add([x1, x2])

        embeddings_initializer = tf.keras.initializers.Constant(embedding_matrix) if embedding_matrix else 'uniform'
        inp = tf.keras.layers.Input(shape=(sequence_length,))
        x = tf.keras.layers.Embedding(
                vocab_size,
                emb_dim,
                input_length=sequence_length,
                embeddings_initializer=embeddings_initializer)(inp)
        x = tf.keras.layers.LayerNormalization()(x)
        for i in range(self.nb_layers):
            x = residual_conv_block(x)
        if self.attention:
            x = residual_attention(x)
        x = tf.keras.layers.GlobalMaxPooling1D()(x)
        x = tf.keras.layers.Dense(32, activation='relu', kernel_regularizer=tf.keras.regularizers.l2(1e-6))(x)
        x = tf.keras.layers.Dropout(self.dropout)(x)
        x = tf.keras.layers.LayerNormalization()(x)

        output_activation = 'sigmoid' if nb_outputs==1 or self.multilabel else 'softmax'
        out = tf.keras.layers.Dense(nb_outputs, activation=output_activation, kernel_regularizer=tf.keras.regularizers.l2(1e-6))(x)
        self.model = tf.keras.Model(inp, out)

        optimizer = tf.keras.optimizers.Adam(lr=self.learning_rate, clipnorm=1.0)
        self.model.compile(optimizer=optimizer, loss='binary_crossentropy', metrics=[])

        X_train, X_val, Y_train, Y_val = train_test_split(X, Y, test_size=0.1, shuffle= True)
        metrics = Metrics(validation_data=(X_val, Y_val))
        self.model.fit(X_train, Y_train, epochs=self.nb_epochs, batch_size=self.batch_size, validation_data=(X_val, Y_val), callbacks=[metrics])
        return self

    def predict(self, X):
        return self.model(X).numpy() > 0.5

    def score(self, X, Y):
        Y_pred = self.predict(X)
        return f1_score(Y_pred, Y, average='micro')



import boto3

import json
from io import BytesIO
import gzip
import os


class PolicyDocumentsDownloader:
    """
    Interact with S3 to get policy document texts
    Input:
        bucket_name: the S3 bucket name to get data from
        dir_path: the directory path within this s3 bucket to get policy data from
    """
    def __init__(self, s3, bucket_name, dir_path):
        
        self.s3 = s3
        self.bucket_name = bucket_name
        self.dir_path = dir_path
        self.pdf_keys = self.get_all_s3_keys()

    def get_all_s3_keys(self):
        """
        https://alexwlchan.net/2017/07/listing-s3-keys/
        Get a list of all keys to look for pdfs in the S3 bucket.
        """
        keys = []

        kwargs = {'Bucket': self.bucket_name}
        while True:
            resp = self.s3.list_objects_v2(**kwargs)
            for obj in resp['Contents']:
                keys.append(obj['Key'])

            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break

        pdf_keys = [k for k in keys if self.dir_path in k]

        return pdf_keys

    def get_hashes(self, word_list=None):
        """
        Get a list of policy document hashes from the S3 location
        Input:
            word_list: a list of words to look for in documents, if None then get hashes for all
        Output:
            hashes: a list of dicts with the file hash and the policy doc source where it is from
        """

        print("Getting hashes for policy documents")
        hashes = []
        for key in self.pdf_keys:
            print("Loading "+key)
            key_name = os.path.split(key)[-1]
            response = self.s3.get_object(Bucket=self.bucket_name, Key=key)
            content = response['Body'].read()
            with gzip.GzipFile(fileobj=BytesIO(content), mode='rb') as fh:
                for line in fh:
                    document = json.loads(line)
                    if document['text']:
                        if word_list:
                            if not any(word.lower() in document['text'].lower() for word in word_list):
                                continue
                        hashes.append({
                            "source": key_name,
                            "file_hash" : document['file_hash']
                        })
                print(str(len(hashes))+" documents")

        return hashes

    def download(self, hash_list=None):
        """
        Download the policy document data from S3
        Input:
            hash_list: a list of hashes to specifically download, if None then download all
        """

        print("Downloading policy documents")
        documents = []
        hashes_found = set() # A checker so we dont download duplicates
        for key in self.pdf_keys:
            print("Loading "+key)
            key_name = os.path.split(key)[-1]
            response = self.s3.get_object(Bucket=self.bucket_name, Key=key)
            content = response['Body'].read()
            with gzip.GzipFile(fileobj=BytesIO(content), mode='rb') as fh:
                for line in fh:
                    document = json.loads(line)
                    if ((document['text']) and (document['file_hash'] not in hashes_found)):
                        if hash_list:
                            if document['file_hash'] not in set(hash_list):
                                continue
                        document["source"] = key_name
                        documents.append(document)
                        hashes_found.add(document['file_hash'])
                            
                print(str(len(documents))+" documents")

        return documents

    def save_json(self, documents, file_name):
        print("Saving data ...")
        with open(file_name, 'w', encoding='utf-8') as output_file:
            for document in documents:
                json.dump(document, output_file) 
                output_file.write("\n")
        print("Number of documents saved: " + str(len(documents)))





# encoding: utf-8
"""
BERT Vectorizer that embeds text using a prertained BERT model
"""

import os
import tarfile

import boto3
from botocore import UNSIGNED
from botocore.client import Config

from transformers import BertModel, BertTokenizer
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import torch

from wellcomeml.ml.constants import MODELS_DIR, MODEL_DISPATCH
from wellcomeml.logger import logger


class BertVectorizer(BaseEstimator, TransformerMixin):
    def __init__(self, pretrained='bert', sentence_embedding='mean_second_to_last'):
        """
        Bert vectorizer parameters

        Args:
            pretrained: A pre-trained model name. Currently 'bert' or 'scibert'
            sentence_embedding: How to embedd a sentence using bert's layers.
            Current options:
            'mean_second_to_last', 'mean_last', 'sum_last' or 'mean_last_four'
            Default: `mean_second_to_last`. If a valid option is not set,
            returns the pooler layer (embedding for the token [CLS])
        """
        self.pretrained = pretrained
        self.sentence_embedding = sentence_embedding

    @classmethod
    def save_transformed(cls, path, X_transformed):
        """Saves transformed embedded vectors"""
        np.save(path, X_transformed)

    @classmethod
    def load_transformed(cls, path):
        """Loads transformed embedded vectors"""
        return np.load(path)

    def bert_embedding(self, x):
        # Max sequence length is 512 for BERT
        if len(x) > 512:
            embedded_a = self.bert_embedding(x[:512])
            embedded_b = self.bert_embedding(x[512:])
            return embedded_a + embedded_b

        tokenized_x = self.tokenizer.tokenize("[CLS] " + x + " [SEP]")
        indexed_tokens = self.tokenizer.convert_tokens_to_ids(tokenized_x)

        tokens_tensor = torch.tensor([indexed_tokens])
        segments_tensor = torch.zeros(tokens_tensor.shape, dtype=torch.long)
        with torch.no_grad():
            output = self.model(tokens_tensor, token_type_ids=segments_tensor)
        last_layer = output[2][-1]
        second_to_last_layer = output[2][-2]

        if self.sentence_embedding == 'mean_second_to_last':
            embedded_x = second_to_last_layer.mean(dim=1)
        elif self.sentence_embedding == 'mean_last':
            embedded_x = last_layer.mean(dim=1)
        elif self.sentence_embedding == 'sum_last':
            embedded_x = last_layer.sum(dim=1)
        elif self.sentence_embedding == "mean_last_four":
            embedded_x = torch.stack(output[2][-4:]).mean(dim=0).mean(dim=1)
        else:
            # Else gives the embbedding for the pooler layer. This can be, for
            # example, fed into a softmax classifier
            embedded_x = output[1]

        return embedded_x.cpu().numpy().flatten()

    def transform(self, X, *_):
        return np.array([self.bert_embedding(x) for x in X])

    def fit(self, *_):
        model_name = 'bert-base-uncased' if self.pretrained == 'bert'\
            else 'scibert_scivocab_uncased'

        # If model_name doesn't exist checks cache and change name to
        # full path
        if model_name == 'scibert_scivocab_uncased':
            model_name = check_cache_and_download(model_name)

        logger.info("Using {} embedding".format(model_name))
        self.model = BertModel.from_pretrained(model_name,
                                               output_hidden_states=True)
        self.tokenizer = BertTokenizer.from_pretrained(model_name)
        self.model.eval()
        return self


def check_cache_and_download(model_name):
    """ Checks if model_name is cached and return complete path"""
    os.makedirs(MODELS_DIR, exist_ok=True)

    model_path = os.path.join(MODELS_DIR, model_name)
    if not os.path.exists(model_path):
        logger.info(f"Could not find model {model_name}. Downloading from S3")

        # The following allows to download from S3 without AWS credentials
        s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        tmp_file = os.path.join(MODELS_DIR,
                                MODEL_DISPATCH[model_name]['file_name'])

        s3.download_file(MODEL_DISPATCH[model_name]['bucket'],
                         MODEL_DISPATCH[model_name]['path'], tmp_file)

        tar = tarfile.open(tmp_file)
        tar.extractall(path=MODELS_DIR)
        tar.close()

        os.remove(tmp_file)

    return model_path



from sklearn.metrics import f1_score, precision_score, recall_score
import tensorflow as tf

class Metrics(tf.keras.callbacks.Callback):
    def __init__(self, validation_data):
        self.validation_data = validation_data

    def on_epoch_end(self, epoch, logs):
        X_val = self.validation_data[0]
        Y_val = self.validation_data[1]

        Y_pred = self.model.predict(X_val) > 0.5
        f1 = round(f1_score(Y_val, Y_pred, average='micro'), 4)
        p = round(precision_score(Y_val, Y_pred, average='micro'), 4)
        r = round(recall_score(Y_val, Y_pred, average='micro'), 4)
        print(f" - val metrics: P {p:.4f} R {r:.4f} F {f1:.4f}")
        return



import os

import setuptools

here = os.path.abspath(os.path.dirname(__file__))

"""
Load data from the__versions__.py module. Change version, etc in
that module, and it will be automatically populated here. This allows us to
access the module version, etc from inside python with

Examples:

    >>> from wellcomeml.common import about
    >>> about['__version__']
    2019.10.0

"""

about = {} # type: dict
version_path = os.path.join(here, 'wellcomeml', '__version__.py')
with open(version_path, 'r') as f:
    exec(f.read(), about)

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name=about['__name__'],
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=about['__url__'],
    license=['__license__'],
    packages=setuptools.find_packages(include=["wellcomeml*"]),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'numpy',
        'pandas',
        'boto3',
        'scikit-learn',
        'spacy==2.2.1',
        'nervaluate',
        'twine'
    ],
    extras_require={
        'deep-learning': [
            'tensorflow',
            'torch',
            'transformers<=2.0.0',
            'spacy-transformers'
        ]
    },
    tests_require=[
        'pytest',
        'pytest-cov'
    ]
)



import random

import spacy
from nervaluate import Evaluator


class SpacyNER:
    """
    Train a spaCy NER classifier on texts annotated using Prodigy
    """

    def __init__(self, n_iter=20, dropout=0.2, output=True):

        self.n_iter = n_iter
        self.dropout = dropout
        self.output = output

    def fit(self, X, y):
        """
        X: a list of sentences,
            e.g. ['Professor Smith said', 'authors Smith and Chen found']
        y: a list of prodigy format spans for each element of X,
            e.g. [{'start': 10, 'end': 15, 'label': 'PERSON'}, {'start': 8, 'end': 13, 'label': 'PERSON'}, {'start':18, 'end':22, 'label':'PERSON'}]
        """

        train_data = list(zip(X, y))

        if 'ner' not in self.nlp_model.pipe_names:
            # If you are training on a blank model
            ner = self.nlp_model.create_pipe('ner')
            self.nlp_model.add_pipe(ner, last=True)
        else:
            ner = self.nlp_model.get_pipe("ner")

        # add labels from data

        for spans in y:
            for span in spans:
                ner.add_label(span['label'])

        other_pipes = [pipe for pipe in self.nlp_model.pipe_names if pipe != 'ner']
        with self.nlp_model.disable_pipes(*other_pipes): # only train NER
            optimizer = self.nlp_model.begin_training()

            for i in range(self.n_iter):
                random.shuffle(train_data)
                losses = {}

                for text, spans in train_data:
                    annotations = self._spans_to_entities(spans)
                    self.nlp_model.update(
                        [text],
                        [annotations],
                        sgd=optimizer,
                        losses=losses,
                        drop=self.dropout
                    )

                if self.output:
                    self._print_output(i, losses['ner'])

        return self.nlp_model

    def _print_output(self, batch_i, loss):

        dash = '-' * 15

        if batch_i == 0:
            print(dash)
            print('{:8s}{:13s}'.format("BATCH |", "LOSS"))
            print(dash)
        print('{:<8d}{:<10.2f}'.format(batch_i, loss))


    def predict(self, text):
        """
        Get entity predictions for a piece of text
        """

        doc = self.nlp_model(text)
        pred_entities = []

        for ent in doc.ents:
            pred_entities.append({
                'start': ent.start_char,
                'end': ent.end_char,
                'label': ent.label_,
            })

        return pred_entities

    def score(self, y_true, y_pred, tags):
        """
        Evaluate the model's performance on the data
        for entity types in the 'tags' list, e.g. ['PERSON', 'LOC']
        """

        evaluator = Evaluator(y_true, y_pred, tags=tags)
        results, results_by_tag = evaluator.evaluate()

        score = {tag: results_by_tag[tag]['partial']['f1'] for tag in tags}
        score["Overall"] = results['partial']['f1']
        return score

    def save(self, file_name):

        self.nlp_model.to_disk(file_name)

    def load(self, model=None):
        """
        model:
            The model's location ('models/model')
            or the name of a spaCy model ("en_core_web_sm")
            None (default): Then load a blank spacy english model
        """

        if model:
            self.nlp_model = spacy.load(model)
        else:
            self.nlp_model = spacy.blank("en")

    def _spans_to_entities(self, spans):
        """
        Convert prodigy spans to spacy entities

        from:

        ```
        [
            {'start': 36, 'end': 46, 'label': 'PERSON'},
            {'start': 48, 'end': 58, 'label': 'PERSON'},
            {'start': 61, 'end': 69, 'label': 'PERSON'},
        ]
        ```

        to:

        ```
        {
            'entities': [
                    [36, 46, 'PERSON'],
                    [48, 58, 'PERSON'],
                    [61, 69, 'PERSON'],
                ]
        }
        ```
        """

        entities = []

        for span in spans:
            entities.append(
                [span['start'], span['end'], span['label']]
            )

        return {'entities': entities}
